/***************************************************************************
                          board.cpp  -  description
                             -------------------
    begin                : Thu Aug 17 2000
    copyright            : (C) 2000 by Waldemar Baraldi
    email                : baraldi@lacasilla.com.ar
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "board.h"
#include "none.h"

#define MAX_SPEED   80

Board::Board():AnimatedCanvas(){
  int i,j;

  for (i=0;i<rows;i++)
    for (j=0;j<columns;j++){
      board[i][j]=new None();
      changes[i][j]=false;
    }
  paint_all=true;
  lps=new List();
  background=NULL;
  flow_row=flow_column=-1;
  entry=exit=NULL;
  actual_output=Void;
  actual_input=Void;
  state=Playing;
  rest_coef=1;
  fixed_coef=1;
  tick_amount=1;
  tank=NULL;
}

Board::~Board(){
  int i,j;
  for (i=0;i<rows;i++)
    for (j=0;j<columns;j++)
      delete board[i][j];
  delete lps;
  if (tank) delete tank;
}

void Board::setStartDelay(int delay){
  if (tank) delete tank;
  tank=new Tank(delay);
  tank->setPos(TankX, TankY);
}

void Board::setEntry(Entry * entry, int row, int column){
  flow_row=row;
  flow_column=column;
  actual_input=Void;
  actual_output=entry->getOutput(Void);
  setPipe(entry, row, column);
}

void Board::setExit(Exit * exit, int row, int column){
  setPipe(exit, row, column);
}

bool Board::isRemovable(int row, int column){
  return board[row][column]->isRemovable();
}

void Board::setPipe(Pipe * pipe, int row, int column){
  changes[row][column]=true;
  if (board[row][column]->isRemovable()){
    delete board[row][column];
    board[row][column]=pipe;
    pipe->setPos(column*pipe->width()+x, row*pipe->height()+y);
  }
}

Pipe * Board::getPipe(int row, int column){
  return board[row][column];
}

BoardState Board::getState(){
  return state;
}

void Board::addPointer(Pointer * p){
  lps->insert(lps->getFirst(), (Pointer *)p);
}

void Board::removePointer(Pointer * p){
  lps->remove(lps->indexOf((Pointer*)p));
}

void Board::movePointer(Pointer *p, int row, int column){
  if (row>=0 && row<rows && column>=0 && column<columns){
    if (p->getRow()>=0 && p->getRow()<rows && p->getColumn()>=0 && p->getColumn()<columns)
      changes[p->getRow()][p->getColumn()]=true;
    p->setRowColumn(row, column);
    p->setMoved(true);
  }
}

void Board::setPos(int x, int y){
  Canvas::setPos(x,y);
  int i,j;
  for (i=0;i<rows;i++)
    for (j=0;j<columns;j++)
      board[i][j]->setPos(j*board[i][j]->width()+x, i*board[i][j]->height()+y);
  paint_all=true;
}

void Board::setSpeed(int amount){
  tick_amount=amount;
}

void Board::setMaxSpeed(){
  tick_amount=MAX_SPEED;
}

void Board::setRestrictionCoef(unsigned int coef){
  rest_coef=coef%101;
}

unsigned int Board::getRestrictionCoef(){
  return rest_coef;
}

void Board::setFixedCoef(unsigned int coef){
  fixed_coef=coef%101;
}

unsigned int Board::getFixedCoef(){
  return fixed_coef;
}

void Board::tick(){
  int aux;

  tank->tick();
  if (!(tank->isEmpty())){//Descuento del tanque
    tank->decFullLevel(tick_amount);
    changes[flow_row][flow_column]=true;
  }
  else{//Aumento el actual
    aux=(getPipe(flow_row, flow_column))->getFullLevel(actual_input);
    if (aux<(getPipe(flow_row, flow_column))->full()){
      (getPipe(flow_row, flow_column))->incFullLevel(actual_input, tick_amount);
      changes[flow_row][flow_column]=true;
    }
    else{//Paso al siguiente
      Pipe * next=NULL;
      actual_input=Void;
      switch (actual_output){
        case North:{
          if (flow_row>0){
            next=getPipe(--flow_row, flow_column);
            actual_input=South;
          }
          break;
        }
        case South:{
          if (flow_row<rows-1){
            next=getPipe(++flow_row, flow_column);
            actual_input=North;
          }
          break;
        }
        case East:{
          if (flow_column<columns-1){
            next=getPipe(flow_row, ++flow_column);
            actual_input=West;
          }
          break;
        }
        case West:{
          if (flow_column>0){
            next=getPipe(flow_row, --flow_column);
            actual_input=East;
          }
          break;
        }
        default:break;
      }
      if (actual_output==Void)
        state=OverWon;
      else
        if (actual_input==Void){
          state=OverLost;
      }else{
        if (next->hasConnection(actual_input) && !(next->isRestrictedAsOutput(actual_input))){
           actual_output=next->getOutput(actual_input);
           next->incFullLevel(actual_input,tick_amount);
           changes[flow_row][flow_column]=true;
         }else{
           state=OverLost;
         }
      }
    }
  }
}

void Board::paint(VideoManager * vm){
  int i,j;
  Index * index;
  Pointer * aux;
  int pc, pr;

  /** Carga el background si es la primera vez que entra.*/
  if (!background)
    background=(vm->getImageManager())->getImage(new Str(Background));

  /** Pinta todos los pipes.*/
  if (paint_all){
    for (i=0;i<rows;i++)
      for (j=0;j<columns;j++){
        vm->blit(background, j*board[i][j]->width()+x, i*board[i][j]->height()+y);
        board[i][j]->paint(vm);
        changes[i][j]=false;
      }
  }else{/** Pinta solo los que han cambiado*/
    for (i=0;i<rows;i++)
      for (j=0;j<columns;j++){
        if (changes[i][j]){
          vm->blit(background, j*board[i][j]->width()+x, i*board[i][j]->height()+y);
          board[i][j]->paint(vm);
          changes[i][j]=false;
        }
      }
  }
  /** Pinta y marca los punteros para la proxima entrada.*/
  index=lps->getFirst();
  for (i=0;i<lps->nObjects();i++){
    aux=(Pointer*)lps->getObject(index);
    pr=aux->getRow();
    pc=aux->getColumn();
    if (aux->moved() || (flow_column==aux->getColumn() && flow_row==aux->getRow())){
      /* No siempre la entrada o la salida se repintan, hay que forzarlo*/
/*      if (flow_column==aux->getColumn() && flow_row==aux->getRow()){
        if (board[flow_row][flow_column]==entry || board[flow_row][flow_column]==exit)
          vm->blit(background, flow_column*board[flow_row][flow_column]->width()+x, flow_row*board[flow_row][flow_column]->height()+y);
          board[flow_row][flow_column]->paint(vm);
      }*/

      aux->setPos(pc*aux->width+x, pr*aux->height+y);
      aux->paint(vm);
      if (aux->moved()) aux->setMoved(false); else aux->setMoved(true);
    }
    index=lps->getNext(index);
  }
  paint_all=false;

  /** Pinta la mancha si el juego esta perdido.*/
  if (state==OverLost){
    Image * mancha=vm->getImageManager()->getImage(new Str("splash.png"));
    int mx=flow_column*board[i][j]->width()+x;
    int my=flow_row*board[i][j]->height()+y;
    switch (actual_input){
      case Void:{ //Esta contra un borde
        switch (actual_output){
          case North:my-=board[i][j]->height()/2;break;
          case South:my+=board[i][j]->height()/2;break;
          case West:mx-=board[i][j]->width()/2;break;
          case East:mx+=board[i][j]->width()/2;break;
          default:break;
        }
        break;
      }
      case North:my-=board[i][j]->height()/2;break;
      case South:my+=board[i][j]->height()/2;break;
      case West:mx-=board[i][j]->width()/2;break;
      case East:mx+=board[i][j]->width()/2;break;
      default:break;
    }
    vm->setClipping(x, y, width(), height());
    vm->enableClipping(true);
    vm->blit(mancha, mx, my);
    vm->enableClipping(false);
  }
  if (tank->isChanged()) tank->paint(vm);
}

int Board::width(){
 return columns*board[0][0]->width();
}

int Board::height(){
 return rows*board[0][0]->height();
}
